c  How do you handle continguous areas in the image processing work?
c  Draw rectangular regions from centroids: min physical size 0.5" square 
c===============================================================================
      program kmeans

c--Group nd data points for nv variables into nz zones. 

c  http://people.sc.fsu.edu/~jburkardt/f_src/kmeans/kmeans.html
c  https://orion.math.iastate.edu/burkardt/f_src/kmeans/kmeans.html

      include 'common.f'
      
c--List of variables c**----- update, all variables from common.
c   kz(nd)     Zone to which data point assigned, output.
c   lz(nz)     Number of data points assigned to a zone, output.
c   c(nz,nv)   Zone centers; input initial guess; updated here.
c   dv(nz)     Sum of squared deviations of variables from zone centers, output.
c   f(nd)      Working array.
c   v(nd,nv)   Variable data, input.
c   nd         Number of data points.
c   nv         Number of variables.
c   nz         Number of zones. 
c   ndmin      Minimum # of data points in a zone (% of total points).
      
c--Read input data.
      call input1
      call input2
      call input3

c--Options for defining zones (kopt):

c  1. Use only geometrical distance (geo).
c  2. Use weighted state variables at end of forging (wsv).
c  3. Use weighted point tracking for state variables all thru forging (ptr).
c     Point track all nodes from last step.
     
c--Output at each iteration.
      open(unit=3,file='iter.dat')

      if(kopt.eq.1)call k1geo
      if(kopt.eq.2)call k2wsv
      if(kopt.eq.3)call k3ptr

      call output
      
      stop; end

c==Input: kmeans for geometrical distance=======================================
      subroutine input1
      include 'common.f'

C** MOVE TO INPUT.DAT FILE
      itmax=50            ! Maximum number of iterations.
      kwkpc=1             ! Workpiece object number (deform simulation).
      data wt/10*0.1/     ! Weighting factors; sum of all factors=1.
      ndmin=10            ! Minimum # of data points in a zone (% of total points).
      kopt=1              ! 1 geometry; 2 state variables; 3 point track.
      nv=2                ! Number of variables.
      nz=6                ! Number of zones.
      nu=0                ! Number of DEFORM user variables.
      nzx=3; nzy=2;       ! Number of zones in x and y for geometrical zoning. 
      nz=nzx*nzy      

c--Read inputs from deform key file at last step.
      open(unit=1,file='step_last.key')

      do 110 ii=1,1000000
      read(1,*,end=120,err=120)x06

c--Nodal coordinates.
      if(x06.eq.'RZ    ')then
      backspace 1; read(1,*)x06,i1
      if(i1.ne.kwkpc)go to 110
      backspace 1; read(1,*)x06,i1,nn
      read(1,*)(j,xn(i),yn(i),i=1,nn)
      endif

c--Element data.
      if(x06.eq.'ELMCON')then
      backspace 1; read(1,*)x06,i1
      if(i1.ne.kwkpc)go to 110
      backspace 1; read(1,*)x06,i1,nel
      read(1,*)(k,(jcon(i,j),j=1,4),i=1,nel)
      endif

  110 continue
  120 close(1)
   
c--Number of data points (elements).
      nd=nel; ndmin=nd/ndmin; 
      
c--Max and min x and y values.
      xmax=maxval(xn(1:nn)); xmin=minval(xn(1:nn))
      ymax=maxval(yn(1:nn)); ymin=minval(yn(1:nn))
      
c--Centers of elements.
      do 130 iel=1,nel
      xe(iel)=0.25*sum(xn(jcon(iel,1:4)))
  130 ye(iel)=0.25*sum(yn(jcon(iel,1:4))) 

      return; end

c==kmeans for geometrical distance==============================================
      subroutine k1geo
      include 'common.f'

c--Initial guess for zone center coordinates.
      iz=0
      do 105 izy=1,2*nzy,2
      do 105 izx=1,2*nzx,2
      iz=iz+1
      cx(iz)=xmin+(xmax-xmin)*real(izx)/real(2*nzx)
  105 cy(iz)=ymin+(ymax-ymin)*real(izy)/real(2*nzy)

      lz(1:nz)=0

c--Distance of each point from zone center; assign to nearest.
      do 115 id=1,nd
      dmin=1.0e10
      do 110 iz=1,nz
      dvc=(xe(id)-cx(iz))**2+(ye(id)-cy(iz))**2
      if(dvc.ge.dmin) go to 110
      dmin=dvc
      f(id)=dmin
      kz(id)=iz
  110 continue
      lz(kz(id))=lz(kz(id))+1
  115 continue

c--New centroids for each zone.
      cx(1:nz)=0.0; cy(1:nz)=0.0;

      do 120 id=1,nd
      cx(kz(id))=cx(kz(id))+xe(id)
  120 cy(kz(id))=cy(kz(id))+ye(id)

      do 125 iz=1,nz
      cx(iz)=cx(iz)/real(lz(iz))
  125 cy(iz)=cy(iz)/real(lz(iz))

c--Check if a point needs to be reassigned to a different zone.
      iter=0
  140 kr=0
            
      do 160 id=1,nd
      ir=kz(id)
c--Bypass iteration if number of zone points less than specified minimum, ndmin.
      if(lz(kz(id)).le.ndmin) go to 160
      dmin=f(id)
      do 155 iz=1,nz
      if(iz.eq.kz(id)) go to 155
      dvc=(xe(id)-cx(iz))**2+(ye(id)-cy(iz))**2
      if(dvc.ge.dmin)go to 155
      dmin=dvc
      ir=iz
  155 continue
      
c--Reassign zone, if necessary.
      if(ir.eq.kz(id))go to 160
      lz(kz(id))=lz(kz(id))-1
      lz(ir)=lz(ir)+1
      kz(id)=ir
      kr=kr+1
  160 continue

c--Update centroids for each zone; bypass if no reassignment done.
      if(kr.eq.0) go to 200
      cxo(1:nz)=cx(1:nz); cyo(1:nz)=cy(1:nz)
      cx(1:nz)=0.0; cy(1:nz)=0.0; dv(1:nz,1:nv)=0.0
 
      do 170 id=1,nd
      cx(kz(id))=cx(kz(id))+xe(id)
  170 cy(kz(id))=cy(kz(id))+ye(id)

c--Change centroid locations gradually to avoid oscillations.
      cdmp=0.5
      if(kr.le.nd/100)cdmp=1.0 
      cdmm=1.0-cdmp
      do 175 iz=1,nz
      cx(iz)=cdmm*cxo(iz)+cdmp*cx(iz)/real(lz(iz))
  175 cy(iz)=cdmm*cyo(iz)+cdmp*cy(iz)/real(lz(iz))

c--Update distances of points from new centroids.
      do 180 id=1,nd
      dvc=(xe(id)-cx(kz(id)))**2+(ye(id)-cy(kz(id)))**2
      f(id)=dvc
  180 dv(kz(id),1)=dv(kz(id),1)+dvc
      devsum=sum(dv(1:nz,1))
      dv(1:nz,1)=dv(1:nz,1)/devsum
      fsum=sum(f(1:nd))
      mzsum=sum(lz(1:nz))
      
      iter=iter+1
      
c--Output at each iteration.
      write(3,901)(iz,cx(iz),cy(iz),iz=1,nz)      
      write(3,902)iter,kr,mzsum,devsum,fsum

c--Stop if no reassignments or max iterations exceeded.
      if(kr.ne.0.and.iter.le.itmax)go to 140

  200 close(3)
  
c--Format statements.
  901 format(i5,1p2e10.3)
  902 format(3i6,1p2e10.3)
  
      return; end 

c==Input: kmeans for state variables============================================
      subroutine input2
      include 'common.f'
      
c--These inputs for kopt=2, state variables kmeans.
      if(kopt.ne.2) return

c--Read inputs from deform key file at last step.
      open(unit=1,file='step_last.key')
      do 110 ii=1,1000000
      read(1,*,end=120,err=120)x06

c--Strain.
      if(x06.eq.'STRAIN')then
      backspace 1; read(1,*)x06,i1,nes; 
      if(i1.eq.kwkpc)read(1,*)(j,strn(j),i=1,nes)
      endif

c--Nodal temperatures.
      if(x06.eq.'NDTMP ')then
      backspace 1; read(1,*)x06,i1,nn;
      if(i1.eq.kwkpc)read(1,*)(j,tem(j),zz,i=1,nn)
      endif

c--Damage.
      if(x06.eq.'DAMAGE')then
      backspace 1; read(1,*)x06,i1,ned;
      if(i1.eq.kwkpc)read(1,*)(j,dmg(j),i=1,ned)
      endif

c--User defined element variables.
      if(x06.eq.'USRELM')then
      backspace 1; read(1,*)x06,i1,neu,zz,neusr; 
      if(i1.eq.kwkpc)read(1,*)(j,(usre(j,k),k=1,neusr),i=1,neu)
      endif

c--User defined nodal variables.
      if(x06.eq.'USRNOD')then
      backspace 1; read(1,*)x06,i1,nnu,zz,nnusr; 
      if(i1.eq.kwkpc)read(1,*)(j,(usrn(j,k),k=1,nnusr),i=1,nnu)
      endif

c**Variables ignored: strain components (x,y,z,shear)
c**Strain rate not saved in KEY file; user variable for max strain rate.

  110 continue
  120 close(1)
      
c--Assign and normalize variables between 0 and 1.

c--Strain.
      vmin=minval(strn(1:nd))
      vmax=maxval(strn(1:nd))
      v(1:nd,1)=(strn(1:nd)-vmin)/(vmax-vmin)

c--Temperature; average from nodes to element centers.
      vmin=minval(tem(1:nd))
      vmax=maxval(tem(1:nd))
      v(1:nd,2)=(0.25*(tem(jcon(1:nd,1))+tem(jcon(1:nd,2))
     x +tem(jcon(1:nd,3))+tem(jcon(1:nd,4)))-vmin)/(vmax-vmin)

c** assign all variables and set weighting factors to zero to deactivate.
c** ks,kt,kd,kue1,kue2,kun1,kun2/1,2,3,4,5,7,8/

      return; end

c==kmeans for state variables===================================================      
      subroutine k2wsv
      include 'common.f'
      
c--Initial guess for zone centers (variables normalized 0 to 1).
      do 103 iv=1,nv
      iz=0
      do 102 izz=1,2*nz,2
      iz=iz+1
  102 c(iz,iv)=real(izz)/real(2*nz)
  103 continue
   
      lz(1:nz)=0
      
c--Deviation of each data point from zone center; assign to nearest.
      do 115 id=1,nd
      dmin=1.0e10
      do 110 iz=1,nz
      dvc=0.0     
      do 105 iv=1,nv
      dvc=dvc+wt(iv)*(v(id,iv)-c(iz,iv))**2
      if(dvc.ge.dmin) go to 110
  105 continue
      dmin=dvc
      f(id)=dmin
      kz(id)=iz
  110 continue
      lz(kz(id))=lz(kz(id))+1
  115 continue

c--Calculate the mean for each zone.
      c(1:nz,1:nv)=0.0; 
      do 120 id=1,nd
      do 120 iv=1,nv
  120 c(kz(id),iv)=c(kz(id),iv)+v(id,iv)

      do 125 iv=1,nv
      do 125 iz=1,nz
  125 c(iz,iv)=c(iz,iv)/real(lz(iz))

c--Check if a data point needs to be reassigned to a different zone.
      iter=0
  140 kr=0

      do 160 id=1,nd
      ir=kz(id)

c--Bypass iteration if number of zone points less than specified minimum, ndmin.
      if(lz(kz(id)).le.ndmin) go to 160
      dmin=f(id)
      do 155 iz=1,nz
      if(iz.eq.kz(id)) go to 155
      dvc=0.0
      do 150 iv=1,nv
      dvc=dvc+wt(iv)*(v(id,iv)-c(iz,iv))**2
      if(dvc.ge.dmin)go to 155
  150 continue
      dmin=dvc
      ir=iz
  155 continue

c--Reassign zone, if necessary.
      if(ir.eq.kz(id))go to 160
      lz(kz(id))=lz(kz(id))-1
      lz(ir)=lz(ir)+1
      kz(id)=ir
      kr=kr+1
  160 continue

c** bypass below if kr=0

      co(1:nz,1:nv)=c(1:nz,1:nv)
      c(1:nz,1:nv)=0.0; dv(1:nz,1:nv)=0.0

      do 166 id=1,nd
      do 166 iv=1,nv
  166 c(kz(id),iv)=c(kz(id),iv)+v(id,iv)

      cdmp=0.5
      if(kr.le.nd/100)cdmp=1.0 
      cdmm=1.0-cdmp
      do 168 iv=1,nv
      do 168 iz=1,nz
  168 c(iz,iv)=cdmm*co(iz,iv)+cdmp*c(iz,iv)/real(lz(iz))

      do 175 id=1,nd
      fsum=0.0
      do 170 iv=1,nv
      dvc=wt(iv)*(v(id,iv)-c(kz(id),iv))**2
      fsum=fsum+dvc
  170 dv(kz(id),iv)=dv(kz(id),iv)+dvc
  175 f(id)=fsum

      devsum=sum(dv(1:nz,1:nv))
      dv(1:nz,1:nv)=dv(1:nz,1:nv)/devsum
      
      iter=iter+1
      
      do 180 iz=1,nz
  180 write(3,999)iz,(c(iz,iv),iv=1,nv)
  999 format(i5,1p10e10.3)

      mzsum=sum(lz(1:nz))
      fsum=sum(f(1:nd))
      write(3,*)iter,kr,mzsum,devsum,fsum

c--Stop if no reassignments during this iteration.
      if(kr.ne.0.and.iter.le.itmax)go to 140

      return; end 
c==Input: kmeans for point tracked state variables==============================
      subroutine input3
      include 'common.f'
      
c--These inputs for kopt=3, point tracking kmeans.
      if(kopt.ne.3) return

c--Find number of points and steps - point tracking data saved by step.
      open(unit=1,file='DEF_PTR.RST')
      npt=0; nst=0
      read(1,951)
      do 115 i=1,10000000
      read(1,952,err=120,end=120)ipt,ist
      npt=max0(ipt,npt)
      if(ipt.eq.1)nst=nst+1
  115 continue
      
c--Read point tracking file; file format:
c   1: Point #, Step #, Time, Stroke-X, Stroke-Y, R/X Coord, Z/Y Coord
c   2: Eff Strain, Eff St-Rate,  Density,  Damage,        Temperature
c   3: X/R Stress, Y/Z Stress,   Z Stress, XY/RZ Stress,  Mean Stress
c   4: Effective Stress, Maximum Principal Stress
c   5: X/R Strain, Y/Z Strain,   Z Strain, XY/RZ Strain,  Max prn strain
c   6: X/R Str-rt, Y/Z Str-rt,   Z Str-rt, XY/RZ Str-rt,  Max prn str-rt

  120 rewind 1; read(1,951)
      do 125 is=1,nst
      do 125 ip=1,npt
  125 read(1,953,end=130,err=130)ipt,ist,xp(ip,is),yp(ip,is),
     x eeff(ip,is),sreff(ip,is),dmgp(ip,is),temp(ip,is),x1           ! c**change to strnp, strtp
  130 close(1)

c--Initial guess for zone centers (variables normalized 0 to 1).
      do 140 iv=1,nv
      iz=0
      do 135 izz=1,2*nz,2
      iz=iz+1
  135 c(iz,iv)=real(izz)/real(2*nz)
  140 continue

c--Assign and normalize variables between 0 and 1.

      do 160 is=1,nst

c--Strain.
      vmin=minval(eeff(1:npt,is))
      vmax=maxval(eeff(1:npt,is))
      do 145 ip=1,npt
  145 vp(ip,is,1)=(eeff(ip,is)-vmin)/(vmax-vmin+1.0e-10)

c--Temperature; average from nodes to element centers.
      vmin=minval(temp(1:npt,is))
      vmax=maxval(temp(1:npt,is))
      do 150 ip=1,npt
  150 vp(ip,is,2)=(temp(ip,is)-vmin)/(vmax-vmin+1.0e-10)
  160 continue

      nd=npt; ndmin=nd/ndmin
      
c** read DEF_PTR_User.RST also; assign all variables and set wt to zero to deactivate.
      
  951 format(/// /// /// /// /// /// //)
  952 format(/i5,i7/////)
  953 format(/i5,i7,39x,2e13.5/12x,2e13.5,13x,2e13.5///12x,e13.5/)

      return; end
c==kmeans for point tracked state variables=====================================
      subroutine k3ptr
      include 'common.f'

c--Initial guess for zone centers (variables normalized 0 to 1).
     
      do 103 is=1,nst
      do 103 iv=1,nv
      iz=0
      vmin=minval(vp(1:nd,is,iv))
      vmax=maxval(vp(1:nd,is,iv))
      do 102 izz=1,2*nz,2
      iz=iz+1
  102 cp(iz,is,iv)=(vmax-vmin)*real(izz)/real(2*nz)   
  103 continue

      do 996 iv=1,2
      do 996 is=1,nst
  996 write(3,*)iv,is,(cp(iz,is,iv),iz=1,nz)

      lz(1:nz)=0
      
c--Deviation of each data point from zone center; assign to nearest.
      do 115 id=1,nd
      dmin=1.0e10
      do 110 iz=1,nz
      dvc=0.0   
      do 105 is=1,nst
      do 105 iv=1,nv
      dvc=dvc+wt(iv)*(vp(id,is,iv)-cp(iz,is,iv))**2      
      if(dvc.ge.dmin) go to 110
  105 continue
      dmin=dvc
      f(id)=dmin
      kz(id)=iz
  110 continue
      lz(kz(id))=lz(kz(id))+1
  115 continue

c--Calculate the mean for each zone.
      cp(1:nz,1:nst,1:nv)=0.0; 
      do 120 id=1,nd
      do 120 is=1,nst
      do 120 iv=1,nv
  120 cp(kz(id),is,iv)=cp(kz(id),is,iv)+vp(id,is,iv)

      do 125 iz=1,nz
      do 125 is=1,nst
      do 125 iv=1,nv
  125 cp(iz,is,iv)=cp(iz,is,iv)/real(lz(iz))

      do 19 iz=1,nz
   19 write(3,999)iz,(cp(iz,nst,iv),iv=1,nv)
   

c--Check if a data point needs to be reassigned to a different zone.
      iter=0
  140 kr=0

      do 160 id=1,nd
      ir=kz(id)

c--Bypass iteration if number of zone points less than specified minimum, ndmin.
      if(lz(kz(id)).le.ndmin) go to 160
      dmin=f(id)
      do 155 iz=1,nz
      if(iz.eq.kz(id)) go to 155
      dvc=0.0
      do 150 is=1,nst
      do 150 iv=1,nv
      dvc=dvc+wt(iv)*(vp(id,is,iv)-cp(iz,is,iv))**2
      if(dvc.ge.dmin)go to 155
  150 continue
      dmin=dvc
      ir=iz
  155 continue

c--Reassign zone, if necessary.
      if(ir.eq.kz(id))go to 160
      lz(kz(id))=lz(kz(id))-1
      lz(ir)=lz(ir)+1
      kz(id)=ir
      kr=kr+1
  160 continue

c** bypass below if kr=0

      cpo(1:nz,1:nst,1:nv)=cp(1:nz,1:nst,1:nv)
      cp(1:nz,1:nst,1:nv)=0.0; dv(1:nz,1:nv)=0.0

      do 166 id=1,nd
      do 166 is=1,nst
      do 166 iv=1,nv
  166 cp(kz(id),is,iv)=cp(kz(id),is,iv)+vp(id,is,iv)

      cdmp=0.5
      if(kr.le.nd/100)cdmp=1.0 
      cdmm=1.0-cdmp
      do 168 iz=1,nz
      do 168 is=1,nst
      do 168 iv=1,nv
  168 cp(iz,is,iv)=cdmm*cpo(iz,is,iv)+cdmp*cp(iz,is,iv)/real(lz(iz))



c-------------------
      do 190 id=1,nd
      do 190 is=1,nst
      dvc=wt(1)*(vp(id,is,1)-cp(kz(id),is,1))**2
      dv(kz(id),1)=dv(kz(id),1)+dvc
  190 continue
      write(3,*)(dv(iz,1),iz=1,nz),sum(dv(1:nz,1))
      do 191 id=1,nd
      do 191 is=1,nst
      dvc=wt(2)*(vp(id,is,2)-cp(kz(id),is,2))**2
      dv(kz(id),2)=dv(kz(id),2)+dvc
  191 continue
      write(3,*)(dv(iz,2),iz=1,nz),sum(dv(1:nz,2))
c** residuals of temperature ~2x of strain- scale all variables so residuals about equal.      
c-------------------------

      do 175 id=1,nd
      fsum=0.0
      do 170 is=1,nst
      do 170 iv=1,nv
      dvc=wt(iv)*(vp(id,is,iv)-cp(kz(id),is,iv))**2
      fsum=fsum+dvc
  170 dv(kz(id),iv)=dv(kz(id),iv)+dvc
  175 f(id)=fsum

      devsum=sum(dv(1:nz,1:nv))
      dv(1:nz,1:nv)=dv(1:nz,1:nv)/devsum
      
      iter=iter+1
      
      do 180 iz=1,nz
  180 write(3,999)iz,(cp(iz,nst,iv),iv=1,nv)
  999 format(i5,1p10e10.3)

      mzsum=sum(lz(1:nz))
      fsum=sum(f(1:nd))
      write(3,*)iter,kr,mzsum,devsum,fsum

c--Stop if no reassignments during this iteration.
      if(kr.ne.0.and.iter.le.itmax)go to 140

      return; end
c==Output data==================================================================
      subroutine output
      include 'common.f'

      open(unit=2,file='mtlgrp.key')
      write(2,*)'MTLGRP 1 ',nd,' 0'
      do 205 id=1,nd
  205 write(2,*)id,kz(id)
      close(2)

c**move to output excel.dat
      if(kopt.ne.3)write(4,990)(id,xe(id),ye(id),kz(id),id=1,nd)
      if(kopt.eq.1)write(4,990)(iz,cx(iz),cy(iz),nz+1,iz=1,nz)
      if(kopt.eq.3)write(4,990)(id,xp(id,nst),yp(id,nst),kz(id),id=1,nd)

  990 format(i5,1p2e10.3,i5)      

      mzsum=sum(lz(1:nz))
      devsum=sum(dv(1:nz,1:nv))
      
      do 450 kk=1,nz
      do 450 iv=1,nv
  450 write(3,*)kk,lz(kk),dv(kk,iv)
      write(3,*)'Total',mzsum,devsum
      close(3)
c** print mean, stdev for each zone for each variable.
      return; end
c===============================================================================
c  do ik = 1, i
c    il = b(ik)
c    fl = e(il)
c      f(ik) = f(ik) * fl / ( fl - 1.0)  c**
c  end do
c!
c  do
c    do ik = 1, i
c      il = b(ik)
c      ir = il
c        fl = e(il)
c        dc = f(ik)
c        do in = 1, n
c          if ( in /= il ) then
c            fm = e(in)                       ! c** why fm used? f stores f*n/(n-1)
c            fm = fm / ( fm + 1.0)
c            de = 0.0D+00
c            do ip = 1, j
c              da = x(ik,ip) - d(in,ip)
c              de = de + da * da * fm
c            end do
c            if ( de < dc ) then
c              dc = de
c              ir = in
c            end if
c        end do
c!  Reassignment is made here if necessary.
c        if ( ir /= il ) then
c          fq = e(ir)
c          e(ir) = e(ir) + 1
c          e(il) = e(il) - 1
c          b(ik) = ir
c          do it = 1, i
c            ij = b(it)
c            if ( ij == il .or. ij == ir ) then
c              f(it) = 0.0D+00
c              do iu = 1, j
c                da = x(it,iu) - d(ij,iu)
c                f(it) = f(it) + da * da
c              end do
c              fl = e(ij)
c              f(it) = f(it) * fl / ( fl - 1.0)   ! c**
c            end if
c          end do
c        end if
c      end if
c    end do
c      